/*
 * test.c -- GPSD client routines for AVLD
 * Author: Michael J. Tubby    Date: 29-JAN-2022
 *
 * ABSTRACT
 *
 * A module to test the GPSD connection and version handling
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <sys/time.h>
#include <gps.h>

#include "xtimer.h"
#include "gpsd.h"


#ifdef DEBUG
#if 0
#define DEBUG_GPSD
#endif
#endif

#define GPSD_READ_TIMEOUT		10000L			/* micro seconds -> 10mS */
#define GPSD_RECONNECT_INTERVAL		5000L			/* milli seconds -> 5S */
#define GPSD_OBS_COUNT			10


/*
 * shared variables
 */
uint32_t gpsd_message_count = 0;
time_t gpsd_start_time = (time_t)0;


/*
 * GPSD's storage for client connections
 */
static struct gps_data_t gpsd_data;

/*
 *  local variables
 */
static int state;					/* GPSD connection state */
static uint32_t gpsd_connect_count = 0;
static uint32_t gpsd_disconnect_count = 0;
static time_t gpsd_connect_time = (time_t)0;		/* timestamp for GPSD connected */
static time_t gpsd_disconnect_time = (time_t)0;		/* timestamp for GPSD connected */
static int gpsd_obs_count = 0;
static xtimer_t gpsd_timer;
static int ending = 0;

static char * state_name[] = {
	"Not Connected",
	"Connected",
	"Retry Wait"
};

static char *gpsd_host = "localhost";
static char *gpsd_port = "2947";



/*
 * change_state() - update the state of the GPDS connection
 */
static void change_state(int new)
{
#ifdef DEBUG_GPSD
	printf("change_state(): %s (%d) -> %s (%d)\n", state_name[state], state, state_name[new], new);
#endif

	/* look for changes in state */
	if (state != new) {
        	
		/* change to disconnected */
		if (new == GPSD_STATE_DISCONNECTED) {
			time(&gpsd_disconnect_time);
			++gpsd_disconnect_count;
		}
        	
        	/* change state to connected */
		if (new == GPSD_STATE_CONNECTED) {
	        	time(&gpsd_connect_time);
	        	++gpsd_connect_count;
		}

#if 0	
		/* commented out as standalone code - nowhere to upcall to */
		/* notify GPS layer of new state */
		gps_notify(new);
#endif 

		/* store new state */
	        state = new;
	}
}


/*
 * process_gpsd_data() -- process GPSD data and extract for GPS components needed
 *
 * The GPSD data frames are large and can be up to 8Kb in size as they may contain
 * satellite info, RTCM info, AIS info and a whole load of stuff we do not need in
 * an AVL system.
 *
 * We check that we have the compnents that we need and that they are valid and
 * copy them over to a muck smaller GPS data object with which we work. 
 *
 */
static void process_gpsd_data(struct gps_data_t * gp)
{
	printf("\n--- GPS State: %s (%d) ---\n", state_name[state], state);
	printf("gp->online.tv_sec:	   %lu\n", gp->online.tv_sec);
	printf("gp->satellites_used:      %d\n", gp->satellites_used);
	printf("gp->satellites_visible:   %d\n", gp->satellites_visible);
	printf("gp->fix.status:           %d\n", gp->fix.status);
	printf("gp->fix.mode:	           %d\n", gp->fix.mode);
	printf("gp->fix.time.tv_sec:      %lu\n", gp->fix.time.tv_sec);
	printf("gp->fix.latitude:         %f\n", gp->fix.latitude);
	printf("gp->fix.longitude:        %f\n", gp->fix.longitude);
	printf("gp->fix.altitude:         %f\n", gp->fix.altitude);
	printf("gp->fix.speed:            %f\n", gp->fix.speed);
	printf("gp->fix.track:	           %f\n", gp->fix.track);
	printf("gp->fix.eph:	           %f\n", gp->fix.eph);
	printf("gp->dop.pdop:	           %f\n", gp->dop.pdop);

	++gpsd_message_count;
}


/*
 * gpsd_init() - initialise the connection to GPSD
 */
int gpsd_init(void)
{
	/* record start-up time */
	time(&gpsd_start_time);

	/* set initial state */
	change_state(GPSD_STATE_DISCONNECTED);

	return 1;
}


/*
 * gpsd_run() - main entry point to run the protocol stack(s)
 */
void gpsd_run(void)
{
	static int rc;
	static xtimer_t gpsd_timer;

	/*
	 * run the GPSD connection
	 */
	switch(state) {
	
		case GPSD_STATE_DISCONNECTED:
			/*
			 * disconnected -> attempt to open a connection to GPSD and set it to streaming mode
			 */
			rc = gps_open(gpsd_host, gpsd_port, &gpsd_data);

			if (rc < 0) {
				/* connect failed - perform retry after waiting for a while */
				fprintf(stderr, "gpsd_run(): call to gps_open() failed - is GPSD running?\n");
				xtimer_start(&gpsd_timer, GPSD_RECONNECT_INTERVAL);
				change_state(GPSD_STATE_RETRY_WAIT);
			} else {
				/* connected okay - set streaming mode and change state */
				(void)gps_stream(&gpsd_data, WATCH_ENABLE | WATCH_JSON | WATCH_DEVICE, NULL);
				change_state(GPSD_STATE_CONNECTED);
			}
			break;
		
		case GPSD_STATE_CONNECTED:
			/*
			 * connected -> attempt to read records from GPSD
			 */
			if (gps_waiting(&gpsd_data, GPSD_READ_TIMEOUT)) {

				/* attempt to read the data from GPSD */
				rc = gps_read(&gpsd_data, NULL, 0);

				if (rc > 0) {
					/*
					 * we have data
					 */

					/* check the GPSD version and API info here */
					if (gpsd_data.set & VERSION_SET) {
						
						fprintf(stderr, "GPSD version: %s API version: %d.%d  App built against API version: %d.%d\n",
							gpsd_data.version.release, gpsd_data.version.proto_major, gpsd_data.version.proto_minor, GPSD_API_MAJOR_VERSION, GPSD_API_MINOR_VERSION);
		
						if (gpsd_data.version.proto_major != GPSD_API_MAJOR_VERSION || gpsd_data.version.proto_minor != GPSD_API_MINOR_VERSION) {
							fprintf(stderr, "GPSD API mis-match ... exiting!\n");
							exit(1);
						}

					} else {
						/* process the rest of the data */
						process_gpsd_data(&gpsd_data);
					}
						
					/* reset the obsolescence count because GPSD is alive */
					gpsd_obs_count = GPSD_OBS_COUNT;

				} else if (rc < 0) {
					/*
					 * read error
					 */
					fprintf(stderr, "gps_read() returned error %d - is GPSD running?\n", rc);
						
					/* close connection to GPSD */
					gps_close(&gpsd_data);
					
					/* start retry timer */
					xtimer_start(&gpsd_timer, GPSD_RECONNECT_INTERVAL);
					
					/* change state to wait for connection */
					change_state(GPSD_STATE_RETRY_WAIT);
					
				} else {
					/*
					 * rc=0 => null read - should never get here?
					 */
				}
				
			}
			break;
		
		case GPSD_STATE_RETRY_WAIT:
			/*
			 * retry waiting -> timer running while we wait before reconnecting
			 */
			if (xtimer_expired(&gpsd_timer)) {
				fprintf(stderr, "GPSD retry timer expired - attemtping to re-connect ...\n");
				change_state(GPSD_STATE_DISCONNECTED);
			}
			break;
	}
}


/*
 * gpsd_second() - housekeeping
 */
void gpsd_second(void)
{
	/* if connected perform obsolescence count and reset on GPSD */
	if (state == GPSD_STATE_CONNECTED && gpsd_obs_count) {
		--gpsd_obs_count;
		if (!gpsd_obs_count) {
			fprintf(stderr, "GPSD obs count expired - connection stalled?\n");
			gps_close(&gpsd_data);
			xtimer_start(&gpsd_timer, GPSD_RECONNECT_INTERVAL);
			change_state(GPSD_STATE_RETRY_WAIT);
		}
	}
}


/*
 * gpsd_close() - finish with GPSD
 */
void gpsd_close(void)
{
	(void)gps_stream(&gpsd_data, WATCH_DISABLE, NULL);
	(void)gps_close(&gpsd_data);
}


