/*
 * timer.c -- Linux version of timer code
 */

#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>

#include "xtimer.h"

/*
 * force timezone to GMT to make it work
 */
static struct timezone tz = {0, 0};


/*
 * time_ticks() : return the time as a "tick count" in milliseconds in
 * an unsigned long. This routine is hardware or operating system specific...
 */
static int64_t time_ticks(void)
{
	static struct timeval tv;

	gettimeofday(&tv, &tz);
	return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}


/*
 * xtimer_clear() : Initialise & clear an individual timer
 */
void xtimer_clear(xtimer_t * t)
{
	t->start_value = 0;
	t->expires_at = 0;
	t->state = TIMER_STOPPED;
}


/*
 * xtimer_start() : Start a specified timer running for a period
 */
void xtimer_start(xtimer_t * t, int64_t val)
{
	t->start_value = time_ticks();
	t->expires_at = t->start_value + val;
	t->state = TIMER_RUNNING;
}


/*
 * xtimer_stop() : Stop a timer
 */
void xtimer_stop(xtimer_t * t)
{
	t->state = TIMER_STOPPED;
}


/*
 * xtimer_running() : Check whether a timer is running
 */
int xtimer_running(xtimer_t * t)
{
	return (t->state == TIMER_RUNNING);
}


/*
 * xtimer_stopped() : Check whether a timer is stopped
 */
int xtimer_stopped(xtimer_t * t)
{
	return (t->state == TIMER_STOPPED);
}


/*
 * xtimer_expired() : Test if the timer has expired. If the timer requested
 * has not been started return false.
 */
int xtimer_expired(xtimer_t * t)
{
	static int64_t current;							/* the current time */

	if (t->state == TIMER_RUNNING) {
		current = time_ticks();

		if (current < t->start_value || current >= t->expires_at) {	/* check for wrap-round */
			t->state = TIMER_EXPIRED;
			return 1;
		}
	}

	return 0;
}


/*
 * xtimer_delay() - blocking wait on a timer
 */
void xtimer_delay(int64_t val)
{
	xtimer_t t;

	xtimer_start(&t, val);
	while (!xtimer_expired(&t))
		;
}


